@@ -4,7 +4,7 @@
 #include <memory>       // for bad_weak_ptr
 #include <type_traits>  // for add_lvalue_reference
 
-#include "cpp11/R.hpp"         // for SEXP, SEXPREC, TYPEOF, R_NilValue, R_C...
+#include "cpp11/R.hpp"         // for SEXP, SEXPREC, R_NilValue
 #include "cpp11/protect.hpp"   // for protect, safe, protect::function
 #include "cpp11/r_bool.hpp"    // for r_bool
 #include "cpp11/r_vector.hpp"  // for type_error
@@ -26,15 +26,15 @@ class external_pointer {
     if (data == nullptr) {
       throw type_error(EXTPTRSXP, NILSXP);
     }
-    if (TYPEOF(data) != EXTPTRSXP) {
-      throw type_error(EXTPTRSXP, TYPEOF(data));
+    if (detail::r_typeof(data) != EXTPTRSXP) {
+      throw type_error(EXTPTRSXP, detail::r_typeof(data));
     }
 
     return data;
   }
 
   static void r_deleter(SEXP p) {
-    if (TYPEOF(p) != EXTPTRSXP) return;
+    if (detail::r_typeof(p) != EXTPTRSXP) return;
 
     T* ptr = static_cast<T*>(R_ExternalPtrAddr(p));
 
